//-------------------------------------------------------------------------------------------------
// TheConsole.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.h"
#include "TheConsole.h"
#include "TheVideoSystem.h"
#include "engineAssert.h"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC TheConsole*		TheConsole::_theConsole				= NULL;
STATIC bool				TheConsole::_isConstructed			= false;
STATIC bool				TheConsole::_isInitialized			= false;


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheConsole::TheConsole()
:
_activeChannels( 0xffffffff ), // list of console print channels active by default
_maxLines( 65536 ),
_numLines( 1 ),
_lastLine( 0 ),
_lineCount( 1 )
{
	/// Make sure the singleton object doesn't get constructed twice
	VALIDATE( !_isConstructed );
	VALIDATE( !_isInitialized );

	/// Set construction complete
	_isConstructed = true;
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheConsole::~TheConsole()
{
	VALIDATE( _isConstructed );
	VALIDATE( !_isInitialized );

	/// Set un-constructed
	_isConstructed = false;
}


//-------------------------------------------------------------------------------------------------
// CreateSingleton
//-------------------------------------------------------------------------------------------------
STATIC TheConsole* TheConsole::CreateSingleton()
{
	VALIDATE( !_isConstructed );
	VALIDATE( !_isInitialized );
	VALIDATE( !_theConsole );

	/// Create the singleton and set the static pointer
	_theConsole = new TheConsole;
	return _theConsole;
}


//-------------------------------------------------------------------------------------------------
// Singleton
//-------------------------------------------------------------------------------------------------
STATIC TheConsole* TheConsole::Singleton()
{
	ASSERT( _isConstructed );
	ASSERT( _theConsole );

	return _theConsole;
}


//-------------------------------------------------------------------------------------------------
// DestroySingleton
//-------------------------------------------------------------------------------------------------
STATIC void TheConsole::DestroySingleton()
{
	VALIDATE( _isConstructed );
	VALIDATE( !_isInitialized );
	VALIDATE( _theConsole );

	/// Destroy the singleton and null the static pointer
	delete _theConsole;
	_theConsole = NULL;
}


//-------------------------------------------------------------------------------------------------
// Init
//-------------------------------------------------------------------------------------------------
void TheConsole::Init()
{
//	CONSOLE->Printf( "::: Initialization : Console (starting)\n" );

	/// Make sure the singleton object doesn't get initialized twice
	VALIDATE( this );
	VALIDATE( _isConstructed );
	VALIDATE( !_isInitialized );

	/// Empty console contents
	_WipeConsole();

	/// Set initialization complete
	_isInitialized = true;
	CONSOLE->Printf( "::: Initialization : Console (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// Shutdown
//-------------------------------------------------------------------------------------------------
void TheConsole::Shutdown()
{
	CONSOLE->Printf( "::: Shutdown : Console (starting)\n" );
	VALIDATE( this );
	VALIDATE( _isConstructed );

	/// De-initialize the system
	_isInitialized = false;
//	CONSOLE->Printf( "::: Shutdown : Console (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// _DebugDumpConsoleLine
//-------------------------------------------------------------------------------------------------
void TheConsole::_DebugDumpConsoleLine( int lineNum, int printLineNum )
{
	DebugPrintf( "%5d)  lineNum = %d, LNRLU = %d : %s", printLineNum, lineNum, _GetLineIndexForPrintLineNum( printLineNum ), _consoleText[ lineNum ].c_str() );
}


//-------------------------------------------------------------------------------------------------
// _DebugDumpConsole
//-------------------------------------------------------------------------------------------------
void TheConsole::_DebugDumpConsole()
{
	DebugPrintf( "\nConsole contents are as follows:\n" );

	/// Check if console buffer has yet to fill up
	if( _numLines < _maxLines )
	{
		for( int i = 0; i < _numLines; i ++ )
		{
			_DebugDumpConsoleLine( i, i );
		}
	}
	else // Console buffer is full
	{
		ASSERT( _numLines == _maxLines );

		/// Determine start and end line numbers
		int endIndex = _lastLine;
		int startIndex = _lastLine + 1;
		if( startIndex >= _maxLines )
		{
			startIndex = 0;
		}

		/// Check if the buffer is wrapped (this is the case most of the time)
		if( startIndex > endIndex )
		{
			int i;

			/// Print lines from <startIndex> to end of buffer
			for( i = startIndex; i < _maxLines; i ++ )
			{
				int numberShyOfEnd = (_maxLines - 1) - i + startIndex;
				int printLineNumber = _lineCount - numberShyOfEnd - 1;
				_DebugDumpConsoleLine( i, printLineNumber );
			}

			/// Print lines from beginning of buffer to <endIndex>
			for( i = 0; i <= endIndex; i ++ )
			{
				int numberShyOfEnd = endIndex - i;
				int printLineNumber = _lineCount - numberShyOfEnd - 1;
				_DebugDumpConsoleLine( i, printLineNumber );
			}
		}
		else
		{
			/// Print lines from beginning of buffer to <endIndex>
			for( int i = 0; i < _maxLines; i ++ )
			{
				int printLineNumOffset = _lineCount - _maxLines;
				int printLineNum = i + printLineNumOffset;
				_DebugDumpConsoleLine( i, printLineNum );
			}
		}
	}

	DebugPrintf( "\n" );
}


//-------------------------------------------------------------------------------------------------
// _WipeConsole
//-------------------------------------------------------------------------------------------------
void TheConsole::_WipeConsole()
{
	_consoleText.clear();
	_consoleText.push_back( "" );
	_numLines = 1;
	_lastLine = 0;

	ASSERT( _numLines == (int) _consoleText.size() );
}


//-------------------------------------------------------------------------------------------------
// _PrintLine
//-------------------------------------------------------------------------------------------------
void TheConsole::_PrintLine( const char* text, int textLength )
{
	string newLine( text, static_cast<size_t>(textLength) );
	/// Check if console buffer has yet to fill up
	if( _numLines < _maxLines )
	{
		_consoleText.push_back( newLine );
		_numLines ++;
		_lastLine ++;
		_lineCount ++;
	}
	else // Console buffer is full; overwrite and check for wrap
	{
		/// Determine the next line number to write to
		int nextLineNum = _lastLine + 1;
		if( nextLineNum >= _maxLines )
		{
			nextLineNum = 0;
		}

		/// Overwrite the previous contents of that line number and advance
		_consoleText[ nextLineNum ] = newLine;
		_lastLine = nextLineNum;
		_lineCount ++;
	}

	DebugPrintf( "CONSOLE: %s\n", newLine.c_str() );
	ASSERT( _numLines == (int) _consoleText.size() );

//	_DebugDumpConsole();
}


//-------------------------------------------------------------------------------------------------
// _PrintChars
//-------------------------------------------------------------------------------------------------
void TheConsole::_PrintChars( const char* text )
{
	/// Break into multiple lines if newline characters are present
	const char* line = text;
	const char* scan;// = strchr( text, '\n' );
	while( *line )
	{
		/// Search for next '\n' character
		scan = strchr( line, '\n' );

		/// Check for early end-of-line (if text does not end with a '\n')
		if( !scan )
		{
			_PrintLine( line, strlen( line ) );
			break;
		}

		/// Determine the length of this line fragment
		int lineLength = (scan - line);

		/// Write it to the console buffer
		_PrintLine( line, lineLength );

		/// Advance to character after '\n'
		line = scan + 1;
	}
}


//-------------------------------------------------------------------------------------------------
// Printf
//-------------------------------------------------------------------------------------------------
void TheConsole::Printf( const char* format, ... )
{
	ASSERT( format );

	/// Silently fail if the default channel is not currently active
//	if( !(_activeChannels & CHAN_DEFAULT) )
//		return;

	/// Allocate a text buffer to hold the args-expanded characters
	const unsigned int	TEXT_BUFFER_SIZE = 2048;
	char				text[ TEXT_BUFFER_SIZE ];
	va_list				args;

	/// Write formatted string into <text>
	va_start( args, format );
	_vsnprintf( text, TEXT_BUFFER_SIZE, format, args );
	va_end( args );
	text[ TEXT_BUFFER_SIZE - 1 ] = '\0';

	_PrintChars( text );
}


//-------------------------------------------------------------------------------------------------
// Printf
//-------------------------------------------------------------------------------------------------
void TheConsole::Printf( bitflags_t channels, const char* format, ... )
{
	ASSERT( format );

	/// Silently fail if this Printf does not go to a currently active console output channel
	if( !(channels & _activeChannels) )
		return;

	/// Allocate a text buffer to hold the args-expanded characters
	const unsigned int	TEXT_BUFFER_SIZE = 2048;
	char				text[ TEXT_BUFFER_SIZE ];
	va_list				args;

	/// Write formatted string into <text>
	va_start( args, format );
	_vsnprintf( text, TEXT_BUFFER_SIZE, format, args );
	va_end( args );
	text[ TEXT_BUFFER_SIZE - 1 ] = '\0';

	_PrintChars( text );
}


//-------------------------------------------------------------------------------------------------
// _GetLineIndexForPrintLineNum
//-------------------------------------------------------------------------------------------------
int TheConsole::_GetLineIndexForPrintLineNum( int printLineNum )
{
	int highestPrintNum = GetHighestPrintLineNumber();
	int lowestPrintNum = GetLowestPrintLineNumber();

	/// Check if <printLineNum> maps to a line number currently in the buffer
	ASSERT( lowestPrintNum <= highestPrintNum );
	ASSERT( printLineNum >= lowestPrintNum );
	ASSERT( printLineNum <= highestPrintNum );

	int printLineNumOffset = highestPrintNum - _lastLine;
	int lineIndex = printLineNum - printLineNumOffset;
	if( lineIndex < 0 )
	{
		lineIndex += _numLines;
	}

	ASSERT( lineIndex >= 0 );
	ASSERT( lineIndex < _numLines );

	return lineIndex;
}


//-------------------------------------------------------------------------------------------------
// GetTextForPrintLineNumber
//-------------------------------------------------------------------------------------------------
const string& TheConsole::GetTextForPrintLineNumber( int printLineNum )
{
	int lineIndex = _GetLineIndexForPrintLineNum( printLineNum );

	return _consoleText[ lineIndex ];
}


//-------------------------------------------------------------------------------------------------
// GetHighestPrintLineNumber
//-------------------------------------------------------------------------------------------------
int TheConsole::GetHighestPrintLineNumber() const
{
	return _lineCount - 1;
}


//-------------------------------------------------------------------------------------------------
// GetLowestPrintLineNumber
//-------------------------------------------------------------------------------------------------
int TheConsole::GetLowestPrintLineNumber() const
{
	int highestPrintNum = GetHighestPrintLineNumber();
	int lowestPrintNum = highestPrintNum - _numLines + 1;
	return lowestPrintNum;
}
