//-------------------------------------------------------------------------------------------------
// TheInputSystem.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.h"
#include "TheInputSystem.h"
#include "TheVideoSystem.h"
#include "TheConsole.h"
#include "TheGameInterfaceAbstract.h"
#include "engineAssert.h"

//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC TheInputSystem*	TheInputSystem::_theInputSystem		= NULL;
STATIC bool				TheInputSystem::_isConstructed		= false;
STATIC bool				TheInputSystem::_isInitialized		= false;


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheInputSystem::TheInputSystem()
:
_mouseDeltaX( 0 ),
_mouseDeltaY( 0 ),
_mousePosX( 0 ),
_mousePosY( 0 ),
//_mouseLastKnownPosX( 0 ),
//_mouseLastKnownPosY( 0 ),
_useWindowsMouseCursor( true ),
_doesAppHaveFocus( true )
{
	/// Make sure the singleton object doesn't get constructed twice
	VALIDATE( !_isConstructed );
	VALIDATE( !_isInitialized );

	/// Initialize complex members
	memset( _isKeyDown, 0, sizeof( _isKeyDown ) );

	/// Set construction complete
	_isConstructed = true;
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheInputSystem::~TheInputSystem()
{
	VALIDATE( _isConstructed );
	VALIDATE( !_isInitialized );

	/// Set un-constructed
	_isConstructed = false;
}


//-------------------------------------------------------------------------------------------------
// CreateSingleton
//-------------------------------------------------------------------------------------------------
STATIC TheInputSystem* TheInputSystem::CreateSingleton()
{
	VALIDATE( !_isConstructed );
	VALIDATE( !_isInitialized );
	VALIDATE( !_theInputSystem );

	/// Create the singleton and set the static pointer
	_theInputSystem = new TheInputSystem;
	return _theInputSystem;
}


//-------------------------------------------------------------------------------------------------
// Singleton
//-------------------------------------------------------------------------------------------------
STATIC TheInputSystem* TheInputSystem::Singleton()
{
	ASSERT( _isConstructed );
	ASSERT( _theInputSystem );

	return _theInputSystem;
}


//-------------------------------------------------------------------------------------------------
// DestroySingleton
//-------------------------------------------------------------------------------------------------
STATIC void TheInputSystem::DestroySingleton()
{
	VALIDATE( _isConstructed );
	VALIDATE( !_isInitialized );
	VALIDATE( _theInputSystem );

	/// Destroy the singleton and null the static pointer
	delete _theInputSystem;
	_theInputSystem = NULL;
}


//-------------------------------------------------------------------------------------------------
// Init
//-------------------------------------------------------------------------------------------------
void TheInputSystem::Init()
{
	CONSOLE->Printf( "::: Initialization : Input System (starting)\n" );

	/// Make sure the singleton object doesn't get initialized twice
	VALIDATE( this );
	VALIDATE( _isConstructed );
	VALIDATE( !_isInitialized );

	/// Set initialization complete
	_isInitialized = true;
	CONSOLE->Printf( "::: Initialization : Input System (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// Shutdown
//-------------------------------------------------------------------------------------------------
void TheInputSystem::Shutdown()
{
	CONSOLE->Printf( "::: Shutdown : Input System (starting)\n" );
	VALIDATE( this );
	VALIDATE( _isConstructed );

	/// Restore the cursor if it was in lock & hide mode
	INPUT->UseWindowsMouseCursor( true );

	/// De-initialize the system
	_isInitialized = false;
	CONSOLE->Printf( "::: Shutdown : Input System (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// IsKeyDown
//-------------------------------------------------------------------------------------------------
const bool TheInputSystem::IsKeyDown( byte_t vKeyCode ) const
{
	ASSERT( this );
	ASSERT( _isConstructed );
	ASSERT( _isInitialized );
	ASSERT( vKeyCode >= 0 && vKeyCode <= 0xff );

	return _isKeyDown[ vKeyCode ];
}


//-------------------------------------------------------------------------------------------------
// DidKeyChangeStateThisFrame
//-------------------------------------------------------------------------------------------------
/*
const bool TheInputSystem::DidKeyChangeStateThisFrame( byte_t vKeyCode ) const
{
	ASSERT( this );
	ASSERT( _isConstructed );
	ASSERT( _isInitialized );
	ASSERT( vKeyCode >= 0 && vKeyCode <= 0xff );

	return _didKeyChangeStateThisFrame[ vKeyCode ];
}
*/


//-------------------------------------------------------------------------------------------------
// ProcessInputEvent
//-------------------------------------------------------------------------------------------------
const bool TheInputSystem::ProcessInputEvent( const UINT messageID, const WPARAM wParam, const LPARAM lParam )
{
	inputEventType_e eventType = INPUT_EVENT_NONE;
	bool wasHandled = false;
	byte_t key = byte_t( wParam );

	/// Determine which type of input event it was
	switch( messageID )
	{
		case WM_KEYDOWN:
			if( HIWORD( lParam ) & KF_REPEAT )
				eventType = INPUT_EVENT_KEYREPEAT;
			else
				eventType = INPUT_EVENT_KEYDOWN;
			break;

		case WM_KEYUP:
			eventType = INPUT_EVENT_KEYUP;
			break;

		case 0x020A: // WM_MOUSEWHEEL
			short highWord = wParam >> 16;
			if( highWord > 0 )
			{
				eventType = INPUT_EVENT_MOUSEWHEEL_UP;
				key = highWord / 120; // WHEEL_DELTA
			}
			else if( highWord < 0 )
			{
				eventType = INPUT_EVENT_MOUSEWHEEL_DOWN;
				key = (-highWord) / 120; // WHEEL_DELTA
			}
			break;
	}

	/// Update the key state table
	if( (eventType == INPUT_EVENT_KEYDOWN) || (eventType == INPUT_EVENT_KEYREPEAT) )
	{
		_isKeyDown[ key ] = true;
	}
	else if( eventType == INPUT_EVENT_KEYUP )
	{
		_isKeyDown[ key ] = false;
	}

	/// Tell the game
	if( eventType != INPUT_EVENT_NONE )
	{
		wasHandled = GAMEINTERFACE->ProcessInputEvent( eventType, key );
	}

	return wasHandled;
}


//-------------------------------------------------------------------------------------------------
// Update
//-------------------------------------------------------------------------------------------------
void TheInputSystem::Update()
{
	/// Check if the game has gone out of focus
	if( !_doesAppHaveFocus )
	{
		/// Report no mouse movement to the game if it doesn't have focus
		_mouseDeltaX = 0;
		_mouseDeltaY = 0;
		return;
	}

	/// Get the windows mouse cursor position
	POINT mousePos;
	::GetCursorPos( &mousePos );

	/// Get movement deltas as compared with last frame
	_mouseDeltaX = mousePos.x - _mousePosX;
	_mouseDeltaY = mousePos.y - _mousePosY;

	if( _useWindowsMouseCursor )
	{
		/// Update internal mouse cursor position
		_mousePosX = mousePos.x;
		_mousePosY = mousePos.y;
	}
	else
	{
		/// Clamp (invisible) mouse back to center of screen to make room for more deltas
		int xScreenMiddle = 160; // FIXME: replace with a VideoSystem method!
		int yScreenMiddle = 120; // FIXME: replace with a VideoSystem method!
		SetMousePosScreenAbsolute( xScreenMiddle, yScreenMiddle );
	}
}


//-------------------------------------------------------------------------------------------------
// UseWindowsMouseCursor
//-------------------------------------------------------------------------------------------------
void TheInputSystem::UseWindowsMouseCursor( const bool enable )
{
	/// Check if the state is already correct
	if( _useWindowsMouseCursor == enable )
		return; // no change

	/// Check if enabling or disabling the windows mouse cursor
	if( enable )
	{
		_useWindowsMouseCursor = true;
		::ShowCursor( TRUE );
	}
	else
	{
		_useWindowsMouseCursor = false;
		::ShowCursor( FALSE );
	}
}


//-------------------------------------------------------------------------------------------------
// GetMouseDeltas
// 
// Sets <mouseDeltaX> and <mouseDeltaY> to the amount of mouse movement for this frame
//-------------------------------------------------------------------------------------------------
void TheInputSystem::GetMouseDeltas( int& mouseDeltaX, int& mouseDeltaY )
{
	mouseDeltaX = _mouseDeltaX;
	mouseDeltaY = _mouseDeltaY;
}


//-------------------------------------------------------------------------------------------------
// GetMouseDeltaX
//-------------------------------------------------------------------------------------------------
int TheInputSystem::GetMouseDeltaX()
{
	return _mouseDeltaX;
}


//-------------------------------------------------------------------------------------------------
// GetMouseDeltaY
//-------------------------------------------------------------------------------------------------
int TheInputSystem::GetMouseDeltaY()
{
	return _mouseDeltaY;
}




//-------------------------------------------------------------------------------------------------
// GetMousePosScreenAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::GetMousePosScreenAbsolute( int& screenMousePosX, int& screenMousePosY ) const
{
	screenMousePosX = _mousePosX;
	screenMousePosY = _mousePosY;
}


//-------------------------------------------------------------------------------------------------
// SetMousePosScreenAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::SetMousePosScreenAbsolute( const int screenMousePosX, const int screenMousePosY )
{
	_mousePosX = screenMousePosX;
	_mousePosY = screenMousePosY;
	::SetCursorPos( _mousePosX, _mousePosY );
}


//-------------------------------------------------------------------------------------------------
// GetMousePosClientAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::GetMousePosClientAbsolute( int& clientMousePosX, int& clientMousePosY ) const
{
	VIDEO->ConvertScreenCoordsToClientCoords( _mousePosX, _mousePosY, clientMousePosX, clientMousePosY );
}


//-------------------------------------------------------------------------------------------------
// SetMousePosClientAbsolute
//-------------------------------------------------------------------------------------------------
void TheInputSystem::SetMousePosClientAbsolute( const int clientMousePosX, const int clientMousePosY )
{
	VIDEO->ConvertClientCoordsToScreenCoords( clientMousePosX, clientMousePosY, _mousePosX, _mousePosY );
	::SetCursorPos( _mousePosX, _mousePosY );
}


//-------------------------------------------------------------------------------------------------
// OnAppGainFocus
//-------------------------------------------------------------------------------------------------
void TheInputSystem::OnAppGainFocus()
{
	_doesAppHaveFocus = true;

	/// Force the cursor back to where it was when we lost focus
	::SetCursorPos( _mousePosX, _mousePosY );
}


//-------------------------------------------------------------------------------------------------
// OnAppLoseFocus
//-------------------------------------------------------------------------------------------------
void TheInputSystem::OnAppLoseFocus()
{
	_doesAppHaveFocus = false;
}


