//-------------------------------------------------------------------------------------------------
// TheSoundSystem.h (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_TheSoundSystem_
#define _include_TheSoundSystem_

//-------------------------------------------------------------------------------------------------
// Easy-access macro
//-------------------------------------------------------------------------------------------------
#define SOUND (TheSoundSystem::Singleton())


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "engineCommon.h"
#include <fmod.h>
#include <fmod_errors.h>


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		TheSoundSystem
//
//		Features:
//			- Play sound with speed, volume, and panning.
//			- Stop an already playing sound.
//			- Change speed, volume, and panning of an already playing sound.
//			- Automated fading of speed, volume, and panning of an already playing sound.
//
//		Volume values are 0.0f to 1.0f, with 1.0f = full volume
//		Panning values are -1.0f to +1.0f, with -1 = left and +1 = right
//		Speed values are positive or negative floats, with 1.0f = normal speed and -1.0f = reverse
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class TheSoundSystem
{
private:
	/// Static members
	static TheSoundSystem* _theSoundSystem;
	static bool _isConstructed;
	static bool _isInitialized;

	/// Private constructors
	TheSoundSystem(); // default constructor

public:

	/// Construction, initialization, and singleton methods
	~TheSoundSystem(); // destructor

	static TheSoundSystem* CreateSingleton();
	static TheSoundSystem* Singleton();
	static void DestroySingleton();

	void Init();
	void Shutdown();
	void Update();

	void AbortOnFmodError();

	//---

	// sounds will never loop unless you explicitly play them that way
	enum
	{
		LOOP_DISABLE	= FSOUND_LOOP_OFF,		// don't loop
		LOOP_NORMAL		= FSOUND_LOOP_NORMAL,	// play forward over and over
		LOOP_PINGPONG	= FSOUND_LOOP_BIDI,		// play forward then reverse then forward...
	};
	typedef int PlaybackLoopMode;

	// Functions for loading/unloading samples and setting global sample parameters.
	void RegisterSample( const int soundNumber, const std::string& name );
	void UnregisterSample( const int soundNumber );
	void UnregisterAllSamples();
	void SetConcurrentPlaybackLimit( const dword_t soundNumber, const byte_t maxPlaybackCount );
	
	// Play/Stop functions.

	void PlaySample(
		const dword_t soundNumber, 
		const dword_t instanceId,
		const PlaybackLoopMode loopMode = LOOP_DISABLE,
		const float volume = 1.0f, 
		const float speed = 1.0f, 
		const float pan = 0.0f );

	void StopSample( const dword_t soundNumber, const dword_t instanceId );
	void StopSample( const dword_t soundNumber );
	void StopPlayingAllSamples();

	// Functions that modify an already-played sound.
	// If you try to modify a sound that is no longer playing, nothing will happen.

	void Pause( const dword_t soundNumber, const dword_t instanceId );
	void Unpause( const dword_t soundNumber, const dword_t instanceId );

	void ChangeVolume( 
		const dword_t soundNumber, 
		const dword_t instanceId, 
		const float targetVolume, 
		const float fadeTimeSeconds = 0.0f );

	void ChangePan( 
		const dword_t soundNumber, 
		const dword_t instanceId, 
		const float targetPan, 
		const float fadeTimeSeconds = 0.0f );

	void ChangeSpeed(
		const dword_t soundNumber, 
		const dword_t instanceId, 
		const float targetSpeed, 
		const float fadeTimeSeconds = 0.0f );

	// Functions for getting information about playback.

	void QuerySampleStatus( 
		const dword_t soundNumber,
		const dword_t instanceId,
		bool& isFinished,
		float& currentLeftVolume,
		float& currentRightVolume,
		float& currentSpeed,
		float& currentPan,
		float& percentPlayed );

	void QueryGlobalStats(
		int& numPlayingSamples,
		int& numLoadedSamples,
		float& cpuUsePercent,
		unsigned int& currentBytesAllocated,
		unsigned int& maxBytesAllocated );

	// The buffer contains 512 floating point values that represent each frequency band's amplitude 
	// for the current FMOD SoundSystem mixing buffer. The range of frequencies covered by the spectrum 
	// is 1 to the nyquist frequency or half of the output rate. So if the output rate is 44100, 
	// then frequencies provided are up to 22050. (entry 511)
	const float* GetSpectrumData512();
	void EnableSpectrumData();
	void DisableSpectrumData();

protected:

	// contains engine data about a sound sample
	struct SampleInfo
	{
		// contains a bit of data about a single playing instance of this sample
		struct PlaybackInstance
		{
			PlaybackLoopMode	loopMode;
			dword_t				id;
			int					fmodChannel;
		};

		// key is expiration time so that finished sounds can be quickly removed from the container
		// use multimap because it can have multiple entries with the same key
		typedef std::multimap< dword_t, PlaybackInstance > PlaybackInstanceContainer; 

		// all the data that is maintained per sample follows; if you add some, be sure to update copy constructor

		FSOUND_SAMPLE*				sample;					// fmod's opaque handle data
		byte_t						currentPlaybackCount;	// how many of these are currently playing
		byte_t						maxPlaybackCount;		// how many can play before oldest instance is restarted instead of new instance starting
		PlaybackInstanceContainer	playbackInstances;		// container for data about actively playing instances of this sample

		// must have copy constructor since the map must be properly copied and not just memcpy'd
		SampleInfo( const SampleInfo& copyme )
		  :	sample					( copyme.sample ),
			currentPlaybackCount	( copyme.currentPlaybackCount ),
			maxPlaybackCount		( copyme.maxPlaybackCount ),
			playbackInstances		( copyme.playbackInstances )
		{
		}

		// must define default constructor for STL
		SampleInfo()
		  :	sample					( 0 ),
			currentPlaybackCount	( 0 ),
			maxPlaybackCount		( 0 )
		{
		}
	};

	// key is sound number
	typedef std::map< dword_t, SampleInfo > SampleContainer;
	
	// container for data about all registered samples
	SampleContainer m_samples;

	// quick utility function for calculating, roughly, at what time a sample will finish playing
	dword_t CalcExpirationTime( const SampleInfo& si, const PlaybackLoopMode loop, const float speed );
};





#endif // _include_TheSoundSystem_
