//-------------------------------------------------------------------------------------------------
// Vector2.h (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_Vector2_
#define _include_Vector2_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "engineCommon.h"
#include "engineMath.h"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		Vector2
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class Vector2
{
public:
	float x; // these are an exception to the "no-public-members" rule of thumb
	float y; // ...for the sake of clarity and brevity in use throughout the code

	static const Vector2 ZERO;
	static const Vector2 ONE;

	/// Construction / Destruction
	~Vector2(); // destructor
	Vector2(); // default constructor
	Vector2( const Vector2& vec ); // copy constructor
	Vector2( const double radians ); // construct unit vector from radian angle
	Vector2( const float x, const float y ); // construct from x,y components

	/// Methods
	const float CalcLengthSquared() const;
	const float CalcLength() const;
	const float Normalize();
	const float SetLength( float newLength );
	void SetXY( const float new_x, const float new_y );
	void GetXY( float& get_x, float& get_y ) const;
	Vector2& Clamp( const Vector2& min, const Vector2& max );
	Vector2& SetFromCosSinRadians( double radians );
	Vector2& SetFromCosSinDegrees( double degrees );
	Vector2& SetFromRadiusCosSinRadians( double radius, double radians );
	Vector2& SetFromRadiusCosSinDegrees( double radius, double degrees );
	double GetDegrees() const;
	double GetRadians() const;
	const float CalcMaxComponent() const;
	const float CalcMinComponent() const;
	const Vector2 Lerp( const Vector2& rhs, const float rhs_fraction ) const;
	const bool IsZero() const;
	const bool IsOne() const;

	/// Operators
	const Vector2& operator = ( const Vector2& rhs );
	const Vector2& operator += ( const Vector2& rhs );
	const Vector2& operator -= ( const Vector2& rhs );
	const Vector2& operator *= ( const float scale );
	const Vector2& operator /= ( const float divisor );

	const Vector2 operator - () const;
	const Vector2 operator - ( const Vector2& rhs ) const;
	const Vector2 operator + ( const Vector2& rhs ) const;
	const Vector2 operator * ( const float scale ) const;
	const float operator | ( const Vector2& rhs ) const; // dot product
	const float operator % ( const Vector2& rhs ) const; // 2d cross product (scalar)

	const float operator [] ( const int index ) const;
	float& operator [] ( const int index );

	const bool operator == ( const Vector2& rhs ) const;
	const bool operator != ( const Vector2& rhs ) const;
};

//-------------------------------------------------------------------------------------------------
// External Operators & Functions
//-------------------------------------------------------------------------------------------------
const Vector2 operator * ( const float scale, const Vector2& rhs );
const bool EpsilonEqual( const Vector2& lhs, const Vector2& rhs, const float epsilon );


//-------------------------------------------------------------------------------------------------
// Vector2::~Vector2()
//-------------------------------------------------------------------------------------------------
inline Vector2::~Vector2()
{
}


//-------------------------------------------------------------------------------------------------
// Vector2::Vector2() - default constructor
//-------------------------------------------------------------------------------------------------
inline Vector2::Vector2()
{
	// do nothing - keep this class as flyweight as possible
}


//-------------------------------------------------------------------------------------------------
// Vector2::Vector2( const Vector2& vec ) - copy constructor
//-------------------------------------------------------------------------------------------------
inline Vector2::Vector2( const Vector2& vec )
:
x( vec.x ),
y( vec.y )
{
}


//-------------------------------------------------------------------------------------------------
// Vector2( float, float )
//-------------------------------------------------------------------------------------------------
inline Vector2::Vector2( const float x, const float y )
:
x( x ),
y( y )
{
}


//-------------------------------------------------------------------------------------------------
// Vector2( double <radians> )
// 
// Construct a unit vector pointing at <radians> angle
//-------------------------------------------------------------------------------------------------
inline Vector2::Vector2( const double radians )
:
x( static_cast<float>( cos( radians ) ) ),
y( static_cast<float>( sin( radians ) ) )
{
}


//-------------------------------------------------------------------------------------------------
// operator =
//-------------------------------------------------------------------------------------------------
inline const Vector2& Vector2::operator = ( const Vector2& rhs )
{
	if( this == &rhs )
		return *this;

	x = rhs.x;
	y = rhs.y;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator +=
//-------------------------------------------------------------------------------------------------
inline const Vector2& Vector2::operator += ( const Vector2& rhs )
{
	x += rhs.x;
	y += rhs.y;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -=
//-------------------------------------------------------------------------------------------------
inline const Vector2& Vector2::operator -= ( const Vector2& rhs )
{
	x -= rhs.x;
	y -= rhs.y;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const Vector2& Vector2::operator *= ( const float scale )
{
	x *= scale;
	y *= scale;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const Vector2& Vector2::operator /= ( const float divisor )
{
	if( divisor )
	{
		x /= divisor;
		y /= divisor;
	}
	else
	{
		x = 0.0f;
		y = 0.0f;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator - (unary minus)
//-------------------------------------------------------------------------------------------------
inline const Vector2 Vector2::operator - () const
{
	return Vector2( -x, -y );
}


//-------------------------------------------------------------------------------------------------
// operator -
//-------------------------------------------------------------------------------------------------
inline const Vector2 Vector2::operator - ( const Vector2& rhs ) const
{
	return( Vector2( x - rhs.x, y - rhs.y ) );
}


//-------------------------------------------------------------------------------------------------
// operator +
//-------------------------------------------------------------------------------------------------
inline const Vector2 Vector2::operator + ( const Vector2& rhs ) const
{
	return( Vector2( x + rhs.x, y + rhs.y ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const Vector2 Vector2::operator * ( const float scale ) const
{
	return( Vector2( x * scale, y * scale ) );
}


//-------------------------------------------------------------------------------------------------
// operator | (dot product)
//-------------------------------------------------------------------------------------------------
inline const float Vector2::operator | ( const Vector2& rhs ) const
{
	return( (x * rhs.x) + (y * rhs.y) );
}


//-------------------------------------------------------------------------------------------------
// operator % (2d scalar cross product)
//-------------------------------------------------------------------------------------------------
inline const float Vector2::operator % ( const Vector2& rhs ) const
{
	return( (x * rhs.y) - (y * rhs.x) );
}


//-------------------------------------------------------------------------------------------------
// operator [] const
//-------------------------------------------------------------------------------------------------
inline const float Vector2::operator [] ( const int index ) const
{
//	ASSERT( ( index >= 0 ) && ( index < 2 ) );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator []
//-------------------------------------------------------------------------------------------------
inline float& Vector2::operator [] ( const int index )
{
//	ASSERT( ( index >= 0 ) && ( index < 2 ) );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator ==
//-------------------------------------------------------------------------------------------------
inline const bool Vector2::operator == ( const Vector2& rhs ) const
{
	return( (x == rhs.x) && (y == rhs.y) );
}


//-------------------------------------------------------------------------------------------------
// operator !=
//-------------------------------------------------------------------------------------------------
inline const bool Vector2::operator != ( const Vector2& rhs ) const
{
	return( (x != rhs.x) || (y != rhs.y) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthSquared
//-------------------------------------------------------------------------------------------------
inline const float Vector2::CalcLengthSquared() const
{
	return( x*x + y*y );
}


//-------------------------------------------------------------------------------------------------
// CalcLength
//-------------------------------------------------------------------------------------------------
inline const float Vector2::CalcLength() const
{
	return( (float) sqrt( x*x + y*y ) );
}


//-------------------------------------------------------------------------------------------------
// Normalize
// 
// FIXME:  consider using a reciprocal square root instead of CalcLength
//	for speed optimization purposes (probably considerably faster than
//	a sqrt and a divide).
//-------------------------------------------------------------------------------------------------
inline const float Vector2::Normalize()
{
	const float length = CalcLength();
	if( length )
	{
		/// Divide each component by the length of the vector
		const float invLength = 1.0f / length;
		x *= invLength;
		y *= invLength;
	}

	return( length );
}


//-------------------------------------------------------------------------------------------------
// SetLength
//-------------------------------------------------------------------------------------------------
inline const float Vector2::SetLength( float newLength )
{
	float oldLength = Normalize();
	(*this) *= newLength;
	return oldLength;
}


//-------------------------------------------------------------------------------------------------
// SetXY
//-------------------------------------------------------------------------------------------------
inline void Vector2::SetXY( const float new_x, const float new_y )
{
	x = new_x;
	y = new_y;
}


//-------------------------------------------------------------------------------------------------
// GetXY
//-------------------------------------------------------------------------------------------------
inline void Vector2::GetXY( float& get_x, float& get_y ) const
{
	get_x = x;
	get_y = y;
}


//-------------------------------------------------------------------------------------------------
// Clamp
//-------------------------------------------------------------------------------------------------
inline Vector2& Vector2::Clamp( const Vector2& min, const Vector2& max )
{
	/// Clamp X component
	if( x > max.x )
		x = max.x;
	else if( x < min.x )
		x = min.x;

	/// Clamp Y component
	if( y > max.y )
		y = max.y;
	else if( y < min.y )
		y = min.y;

	return *this;
}


//-------------------------------------------------------------------------------------------------
// SetFromCosSinRadians
//-------------------------------------------------------------------------------------------------
inline Vector2& Vector2::SetFromCosSinRadians( double radians )
{
	x = static_cast<float>( cos( radians ) );
	y = static_cast<float>( sin( radians ) );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// SetFromCosSinDegrees
//-------------------------------------------------------------------------------------------------
inline Vector2& Vector2::SetFromCosSinDegrees( double degrees )
{
	double radians = DEG2RAD( degrees );
	x = static_cast<float>( cos( radians ) );
	y = static_cast<float>( sin( radians ) );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// SetFromRadiusCosSinRadians
//-------------------------------------------------------------------------------------------------
inline Vector2& Vector2::SetFromRadiusCosSinRadians( double radius, double radians )
{
	x = static_cast<float>( radius * cos( radians ) );
	y = static_cast<float>( radius * sin( radians ) );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// SetFromRadiusCosSinDegrees
//-------------------------------------------------------------------------------------------------
inline Vector2& Vector2::SetFromRadiusCosSinDegrees( double radius, double degrees )
{
	double radians = DEG2RAD( degrees );
	x = static_cast<float>( radius * cos( radians ) );
	y = static_cast<float>( radius * sin( radians ) );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// GetDegrees
//-------------------------------------------------------------------------------------------------
inline double Vector2::GetDegrees() const
{
	return RAD2DEG( atan2( y, x ) );
}


//-------------------------------------------------------------------------------------------------
// GetRadians
//-------------------------------------------------------------------------------------------------
inline double Vector2::GetRadians() const
{
	return atan2( y, x );
}


//-------------------------------------------------------------------------------------------------
// CalcMaxComponent
//-------------------------------------------------------------------------------------------------
inline const float Vector2::CalcMaxComponent() const
{
	/// Assume x is the largest
	float max = x;

	/// Check if y is bigger
	if( y > max )
		max = y;

	return( max );
}


//-------------------------------------------------------------------------------------------------
// CalcMinComponent
//-------------------------------------------------------------------------------------------------
inline const float Vector2::CalcMinComponent() const
{
	/// Assume x is the smallest
	float min = x;

	/// Check if y is smaller
	if( y < min )
		min = y;

	return( min );
}


//-------------------------------------------------------------------------------------------------
// Lerp
//-------------------------------------------------------------------------------------------------
inline const Vector2 Vector2::Lerp( const Vector2& rhs, const float rhs_fraction ) const
{
	return Vector2( x + rhs_fraction * (rhs.x - x),
					y + rhs_fraction * (rhs.y - y));
}


//-------------------------------------------------------------------------------------------------
// operator * (scale, Vector2)
//-------------------------------------------------------------------------------------------------
inline const Vector2 operator * ( const float scale, const Vector2& rhs )
{
	return( Vector2( scale * rhs.x, scale * rhs.y ) );
}


//-------------------------------------------------------------------------------------------------
// EpsilonEqual
//-------------------------------------------------------------------------------------------------
inline const bool EpsilonEqual( const Vector2& lhs, const Vector2& rhs, const float epsilon )
{
	return( fabs( lhs.x - rhs.x ) < epsilon &&
			fabs( lhs.y - rhs.y ) < epsilon );
}


//-------------------------------------------------------------------------------------------------
// IsZero
//-------------------------------------------------------------------------------------------------
inline const bool Vector2::IsZero() const
{
	if( (x != 0.0f) || (y != 0.0f) )
		return false;
	else
		return true;
}


//-------------------------------------------------------------------------------------------------
// IsOne
//-------------------------------------------------------------------------------------------------
inline const bool Vector2::IsOne() const
{
	if( (x != 1.0f) || (y != 1.0f) )
		return false;
	else
		return true;
}


#endif // _include_Vector2_
