// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "gameCommon.h"
#include "GameRenderer.h"
#include "rgba.h"
#include "EnemyGuard.h"
#include "TheGame.h"
#include "Player.h"
#include "MapData.h"

GameRenderer g_renderer;

//-------------------------------------------------------------------------------------------------
void GameRenderer::SetBackgroundTexture( unsigned char* pixels, int width, int height )
{
	ASSERT( pixels );

	/// Create an opengl texture name (number) for this image
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 ); // pixel info is byte-aligned
	glGenTextures( 1, &m_texnumBackgroundTexture); // create a texnum for this new image
	glBindTexture( GL_TEXTURE_2D, m_texnumBackgroundTexture ); // select this as the current 2d texture
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

	unsigned char* buffer = new unsigned char[ 2048*2048*3 ];
	memset( buffer, 0xffffffff, 2048*2048*3 );
	for( int y = 0; y < height; y++ )
	{
		for( int x = 0; x < width; x++ )
		{
			memcpy( &buffer[ (3*x) + (y*3*2048) ], &pixels[ (3*x) + (y*3*width) ], 3 );
		}
		
	}

	/// Upload image data to OpenGL
	glTexImage2D( 
		GL_TEXTURE_2D,		// creating a 2d texture
		0,					// use the root picmip level for this texture (0 = best)
		GL_RGB,				// format opengl should use internally for this image
		2048,				// width of the source data, in pixels
		2048,				// height of the source data, in pixels
		0,					// do not add a border
		GL_BGR_EXT,			// the layout of the pixels in the buffer
		GL_UNSIGNED_BYTE,	// pixel color components are unsigned bytes
		buffer );			// buffer containing pixel data bytes

	delete [] buffer;
}

//-------------------------------------------------------------------------------------------------
void GameRenderer::RenderBackground()
{
	const float bottom = 960.0f;
	const float right = 1280.0f;

	glPushAttrib( GL_ENABLE_BIT | GL_LIGHTING_BIT );
	glEnable( GL_TEXTURE_2D );
	glShadeModel( GL_FLAT );
	RENDERER->SetCurrentTexture( m_texnumBackgroundTexture );
	RENDERER->SetColor( Rgba::WHITE );
	glBegin( GL_QUADS );
	{
		const float texRight = right / 2048.0f;
		const float texBottom = bottom / 2048.0f;
		glTexCoord2f( 0.0f, 0.0f ); // bottom-left
		glVertex2f( 0.0f, 0.0f );
		
		glTexCoord2f( 0.0f, texBottom ); // top-left
		glVertex2f( 0.0f, bottom );
		
		glTexCoord2f( texRight, texBottom ); // top-right
		glVertex2f( right, bottom );
		
		glTexCoord2f( texRight, 0.0f ); // bottom-right
		glVertex2f( right, 0.0f );
	}
	glEnd(); // GL_QUADS
	glPopAttrib();

}

//-------------------------------------------------------------------------------------------------
void GameRenderer::RenderEnemies()
{
	/// Draw all enemy guards
	std::list< EnemyGuard* >::iterator iter;
	std::list< EnemyGuard* >& guardList = EnemyGuard::m_guardList;
	for( iter = guardList.begin(); iter != guardList.end(); iter ++ )
	{
		EnemyGuard* const g = *iter;
		g->Render();
	}
}

//-------------------------------------------------------------------------------------------------
void GameRenderer::RenderVisionCone( Vector2& origin, const float forwardDeg, const float visionDist, const float fov, const bool highlight )
{
	Rgba startColor( 1.0f, 0.0f, 0.0f, 1.0f );
	Rgba endColor( 1.0f, 1.0f, 1.0f, 0.1f );
	
	const double halfFov = fov / 2.0;
	double degInc = fov / 30.0;
	Vector2 p;

	// render full cone------------
	glBegin( GL_LINE_LOOP );
	glColor4f( 1.0f, 1.0f, 1.0f, 0.3f );
	glVertex2f( origin.x, origin.y );
	for( double d = forwardDeg - halfFov; d <= forwardDeg + halfFov; d += degInc )
	{
		p.SetFromCosSinDegrees( d );
		p *= visionDist;
		p += origin;
		glVertex2f( p.x, p.y );
	}
	glEnd();

	// render vision------------
	glBegin( GL_TRIANGLE_FAN );
	
	// center color
	if( highlight )
		glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );
	else
		glColor4f( startColor.r, startColor.g, startColor.b, startColor.a );

	// center vertex
	glVertex2f( origin.x, origin.y );

	// fan circle
	for( d = forwardDeg - halfFov; d <= forwardDeg + halfFov; d += degInc )
	{
		p.SetFromCosSinDegrees( d );
		p *= visionDist;
		p += origin;

		int x, y;
		GridDataType type;
		float scanFraction;
		g_mapData->TraceLine( 1, origin, p, x, y, type, scanFraction );

		// fan color
		Rgba c;
		if( highlight )
		{
			c.SetRGBA( 1.0f, 1.0f, 1.0f, 1.0f );
		}
		else
		{
			c = startColor.Lerp( endColor, scanFraction );
		}
		glColor4f( c.r, c.g, c.b, c.a );
		glVertex2i( x, y );
	}
	
	glEnd();
}

//-------------------------------------------------------------------------------------------------
void GameRenderer::RenderBustaculationMeter()
{
	const float right = 1280*g_player->m_bustaculation;
	glBegin( GL_QUADS );
	glColor4f( 1.0f, 0.0f, 0.0f, 1.0f );
	glVertex2f( 0, 0 );
	glVertex2f( right, 0 );
	glVertex2f( right, 32 );
	glVertex2f( 0, 32 );
	glEnd();
}
