// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "gameCommon.h"
#include "Image.h"

const byte_t TGA_HEADER_BYTES[ 12 ] = {0,0,2,0,0,0,0,0,0,0,0,0};

bool LoadTarga( const char* const filename, unsigned char*& pixels, short& width, short& height, int& bytesPerPixel )
{
	FILE* file = fopen( filename, "rb" );
	if( !file )
	{
		VALIDATE_MSG1( file, "Failed to open %s", filename );
		return false;
	}

	/// Read in the 12-byte primer and validate it
	byte_t primerTGA[ 12 ];
	int bytesRead = fread( primerTGA, 1, sizeof( primerTGA ), file );
	VALIDATE( bytesRead == sizeof( primerTGA ) );
	VALIDATE( sizeof( primerTGA ) == sizeof( TGA_HEADER_BYTES ) ); 
	VALIDATE( !memcmp( primerTGA, TGA_HEADER_BYTES, sizeof( primerTGA ) ) );

	/// Read in the 6-byte header core
	byte_t dummyByte;
	fread( &width,				sizeof( width ),			1, file ); // short
	fread( &height,				sizeof( height ),			1, file ); // short
	fread( &bytesPerPixel,		sizeof( bytesPerPixel ),	1, file ); // byte
	fread( &dummyByte,			sizeof( dummyByte ),		1, file ); // byte
	if( bytesPerPixel != 32 && bytesPerPixel != 24 )
	{
		PANIC_MSG2( "LoadTarga : %s had unsupported bit-depth of %d bpp\n", filename, bytesPerPixel );
		return false;
	}

	bytesPerPixel /= 8;

	/// Calculate pixel data size
	const int pixelBytes = width * height * bytesPerPixel;

	/// Allocate memory for the data segment
	pixels = new byte_t[ pixelBytes ];

	/// Read in the data segment (pixels in BGR format)
	// read in backwards
	const int rowSize = width * bytesPerPixel;
	for( int y = height - 1; y >= 0 ; --y )
	{
		bytesRead = fread( &pixels[ y * rowSize ], sizeof( byte_t ), rowSize, file ); // read in BGR data
		VALIDATE( bytesRead == rowSize );
	}
	fclose( file );

	return true;
}


/*
//-------------------------------------------------------------------------------------------------
Image::ImageMap Image::m_images;
byte_t Image::TGA_HEADER_BYTES[ 12 ] = {0,0,2,0,0,0,0,0,0,0,0,0};

//-------------------------------------------------------------------------------------------------
Image::Image( const std::string& filename )
:	m_glTexNum		( 0 ),
	m_width			( 0 ),
	m_height		( 0 ),
	m_pixels		( 0 )
{
	Load( filename );
}

//-------------------------------------------------------------------------------------------------
Image::~Image()
{
	if( m_glTexNum )
	{
		glDeleteTextures( 1, &m_glTexNum );
	}
	if( m_pixels )
	{
		delete [] m_pixels;
	}
}

//-------------------------------------------------------------------------------------------------
void Image::Load( const std::string& filename )
{
	// reset handles and filename
	if( m_glTexNum )
	{
		glDeleteTextures( 1, &m_glTexNum );
		m_glTexNum = 0;
	}
	m_filename = filename;

	FILE* file = fopen( filename.c_str(), "rb" );
	VALIDATE_MSG1( file, "Failed to open %s", filename.c_str() );

	/// Read in the 12-byte primer and validate it
	byte_t primerTGA[ 12 ];
	int bytesRead = fread( primerTGA, 1, sizeof( primerTGA ), file );
	VALIDATE( bytesRead == sizeof( primerTGA ) );
	VALIDATE( sizeof( primerTGA ) == sizeof( TGA_HEADER_BYTES ) ); 
	VALIDATE( !memcmp( primerTGA, TGA_HEADER_BYTES, sizeof( primerTGA ) ) );

	/// Read in the 6-byte header core
	byte_t dummyByte;
	fread( &m_width,			sizeof( m_width ),			1, file ); // short
	fread( &m_height,			sizeof( m_height ),			1, file ); // short
	fread( &m_bitsPerPixel,		sizeof( m_bitsPerPixel ),	1, file ); // byte
	fread( &dummyByte,			sizeof( dummyByte ),		1, file ); // byte

	/// Set the proper opengl image format
	if( m_bitsPerPixel == 32 )
	{
		m_internalFormat = GL_RGBA;
		m_bufferFormat = GL_BGRA_EXT;
	}
	else if( m_bitsPerPixel == 24 )
	{
		m_internalFormat = GL_RGB;
		m_bufferFormat = GL_BGR_EXT;
	}
	else
	{
		PANIC_MSG2( "Image::_LoadTGAFromFile : %s had unsupported bit-depth of %d bpp\n", filename.c_str(), m_bitsPerPixel );
	}

	/// Calculate pixel data size and <bytesPerPixel>
	m_bytesPerPixel = m_bitsPerPixel / 8;
	const int pixelBytes = m_width * m_height * m_bytesPerPixel;

	/// Allocate memory for the data segment
	m_pixels = new byte_t[ pixelBytes ];

	/// Read in the data segment (pixels in BGR format)
	bytesRead = fread( m_pixels, sizeof( byte_t ), pixelBytes, file ); // read in BGR data
	VALIDATE( bytesRead == pixelBytes );
	fclose( file );
}

//-------------------------------------------------------------------------------------------------
inline int CountBits( int value )
{
	byte count = 0;
	while( value )
	{
		if( value & 1 ) count++;
		value >>= 1;
	}
	return count;
}

//-------------------------------------------------------------------------------------------------
void Image::MakeGLTexture()
{
	ASSERT( m_pixels );
	VALIDATE_MSG1( CountBits( m_width ) == 1, "Width for %s is not power of 2!", m_filename.c_str() );
	VALIDATE_MSG1( CountBits( m_height ) == 1, "Height for %s is not power of 2!", m_filename.c_str() );

	/// Create an opengl texture name (number) for this image
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 ); // pixel info is byte-aligned
	glGenTextures( 1, &m_glTexNum); // create a texnum for this new image
	glBindTexture( GL_TEXTURE_2D, m_glTexNum ); // select this as the current 2d texture
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

	/// Upload image data to OpenGL
	glTexImage2D( 
		GL_TEXTURE_2D,		// creating a 2d texture
		0,					// use the root picmip level for this texture (0 = best)
		m_internalFormat,	// format opengl should use internally for this image
		m_width,				// width of the source data, in pixels
		m_height,			// height of the source data, in pixels
		0,					// do not add a border
		m_bufferFormat,		// the layout of the pixels in the buffer
		GL_UNSIGNED_BYTE,	// pixel color components are unsigned bytes
		m_pixels );		// buffer containing pixel data bytes
};


//-------------------------------------------------------------------------------------------------
// STATIC 
Image& Image::Create( const int userHandle, const std::string& filename )
{
	Image* newImage = new Image( filename );
	
	// don't allow loading the same handle twice
	ASSERT( m_images.find( userHandle ) == m_images.end() );

	m_images[ userHandle ] = newImage;
	return *newImage;
}

//-------------------------------------------------------------------------------------------------
// STATIC 
void Image::Destroy( const int userHandle )
{
	ImageMap::iterator i = m_images.find( userHandle );
	ASSERT( i != m_images.end() );
	
	// delete the image and remove it from the cache
	delete i->second;
	m_images.erase( i );
}


//-------------------------------------------------------------------------------------------------
// STATIC 
void Image::Draw( const Vector2& centerPoint, float scale, const Rgba& rgba, bool setOpenglStates ) const
{
	Vector2 rectSize = scale * GetSizeVector();
	RENDERER->DrawTexturedRectCentered( m_glTexNum, centerPoint, rectSize, rgba, setOpenglStates );
}





*/