// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.h"
#include "TheGame.h"
#include "Player.h"
#include "MapData.h"
#include "EnemyGuard.h"


//-------------------------------------------------------------------------------------------------
// constants
//-------------------------------------------------------------------------------------------------



Player* g_player = NULL;
bool g_collisionFlag = false;


//-------------------------------------------------------------------------------------------------
// LocatePlayerStartPosition
//-------------------------------------------------------------------------------------------------
void Player::LocatePlayerStartPosition()
{
	for( int x = 0; x < MAP_WIDTH; x ++ )
	for( int y = 0; y < MAP_HEIGHT; y ++ )
	{
		if( g_mapData->data[ x ][ y ].type == GRIDTYPE_PLAYERSTART )
		{
			m_origin.SetXY( (float)x, (float)y );
			return;
		}
	}

	PANIC_MSG3( "Could not find a player start (rgb=%d,%d,%d) position on the map!\n", COLOR_PLAYERSTART.r, COLOR_PLAYERSTART.g, COLOR_PLAYERSTART.b );
}


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
Player::Player()
:
m_origin( 100.0f, 100.0f ),
m_forwardDeg( 0.0f ),
m_fov( 30.0f ),
m_visionDist( 32.0f ),
m_movingRight( false ),
m_movingUp( false ),
m_movingLeft( false ),
m_movingDown( false )
{
	/// Find the player start spot and put him there
	LocatePlayerStartPosition();
}


//-------------------------------------------------------------------------------------------------
// Draw
//-------------------------------------------------------------------------------------------------
void Player::Draw()
{
	float fRadius = (float) PLAYER_RADIUS;

	RENDERER->SetColor( Rgba::GREEN );
	if( g_collisionFlag )
	{
		RENDERER->SetColor( Rgba::RED );
	}

	glBegin( GL_QUADS );
	{
		glVertex2f( m_origin.x - fRadius, m_origin.y - fRadius );
		glVertex2f( m_origin.x + fRadius, m_origin.y - fRadius );
		glVertex2f( m_origin.x + fRadius, m_origin.y + fRadius );
		glVertex2f( m_origin.x - fRadius, m_origin.y + fRadius );
	}
	glEnd(); // GL_QUADS
}


//-------------------------------------------------------------------------------------------------
// OnKeyAction
//-------------------------------------------------------------------------------------------------
void Player::OnKeyAction( const byte_t key, bool isKeyDown )
{
	if( key == VK_RIGHT )
	{
		m_movingRight = isKeyDown;
	}
	else if( key == VK_UP )
	{
		m_movingUp = isKeyDown;
	}
	else if( key == VK_LEFT )
	{
		m_movingLeft = isKeyDown;
	}
	else if( key == VK_DOWN )
	{
		m_movingDown = isKeyDown;
	}
}


//-------------------------------------------------------------------------------------------------
// Move
//-------------------------------------------------------------------------------------------------
void Player::Move()
{
	/// Calculate the move vector for this frame
	const float moveSpeed = PLAYER_MOVE_SPEED;
	Vector2 moveVector( Vector2::ZERO );

	/// Build time-independent movement vector (movement per second)
	if( m_movingUp )
		moveVector += Vector2( 0.0f, -1.0f );

	if( m_movingDown )
		moveVector += Vector2( 0.0f, 1.0f );

	if( m_movingRight )
		moveVector += Vector2( 1.0f, 0.0f );

	if( m_movingLeft )
		moveVector += Vector2( -1.0f, 0.0f );

	/// Normalize (prevents strafe-running)
	moveVector.Normalize();
	moveVector *= PLAYER_MOVE_SPEED;

	/// Scale by time
	moveVector *= (float) APPLICATION->GetFrameSeconds();
	Vector2 newOrigin = m_origin + moveVector;

	/// Set forward degrees and forward unit vector
	m_forwardDeg = (float) moveVector.GetDegrees();
	m_forward.SetFromCosSinDegrees( m_forwardDeg );

	/// Check collision on this proposed move
	bool collision = DoesMovementCauseCollision( m_origin, newOrigin );

	/// Update the player's position
	g_collisionFlag = collision;
	if( !collision )
	{
		m_origin = newOrigin; 
	}
}


//-------------------------------------------------------------------------------------------------
// DoesMovementCauseCollision
//-------------------------------------------------------------------------------------------------
bool Player::DoesMovementCauseCollision( Vector2& start, Vector2& end )
{
	VALIDATE( g_mapData );

	/// Get integral versions of current and new position(s)
	int startX = (int) start.x;
	int startY = (int) start.y;
	int endX = (int) end.x;
	int endY = (int) end.y;

	/// Get integral mins and maxs of player's new proposed moved bounds
	int endMinX = endX - PLAYER_RADIUS;
	int endMaxX = endX + PLAYER_RADIUS;
	int endMinY = endY - PLAYER_RADIUS;
	int endMaxY = endY + PLAYER_RADIUS;

	/// Check each pixel in the map grid to see if there is collision 
	for( int x = endMinX; x <= endMaxX; x ++ )
	for( int y = endMinY; y <= endMaxY; y ++ )
	{
		if( g_mapData->data[ x ][ y ].type == GRIDTYPE_WALL )
		{
			return true;
		}
	}

	return false;
}


//-------------------------------------------------------------------------------------------------
void Player::UpdateBustaculation()
{
	// count bustaculations
	std::list< EnemyGuard* >::iterator iter;
	std::list< EnemyGuard* >& guardList = EnemyGuard::m_guardList;
	for( iter = guardList.begin(); iter != guardList.end(); iter ++ )
	{
		const EnemyGuard* const e = *iter;
		if( e->m_canSeePlayer )
		{
			m_bustaculation += BUSTACULATION_PER_SECOND * (float)APPLICATION->GetFrameSeconds();
			if( m_bustaculation > 1.0f )
				m_bustaculation = 1.0f;

			return;
		}
	}

	m_bustaculation -= DISBUSTACULATION_PER_SECOND * (float)APPLICATION->GetFrameSeconds();
	if( m_bustaculation < 0 )
		m_bustaculation = 0;
}