//-------------------------------------------------------------------------------------------------
// TheGame.cpp (game)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.h"
#include "TheGame.h"
#include "Image.h"
#include "MapData.h"
#include "GameRenderer.h"
#include "Player.h"
#include "EnemyGuard.h"
#include "GameBalance.h"


//-------------------------------------------------------------------------------------------------
// Globals
//-------------------------------------------------------------------------------------------------
TheGame* g_game = NULL;


//-------------------------------------------------------------------------------------------------
// Destructor()
//-------------------------------------------------------------------------------------------------
TheGame::~TheGame()
{
}


//-------------------------------------------------------------------------------------------------
// Constructor
//-------------------------------------------------------------------------------------------------
TheGame::TheGame()
{
}


//-------------------------------------------------------------------------------------------------
// PreInit
//-------------------------------------------------------------------------------------------------
void TheGame::PreInit()
{
	/// Assign application name
	APPLICATION->SetName( "Bustaculated" );

	/// Choose a video mode
//	VIDEO->SetDefaultVidModeProperties( 1280, 960, 32, 85, TheVideoSystem::VIDMODE_FLAG_FULLSCREEN );
	VIDEO->SetDefaultVidModeProperties( 1280, 960, 32, 85, 0 );

	/// Set the mouse in hide & recenter mode
//	INPUT->UseWindowsMouseCursor( false );

}


//-------------------------------------------------------------------------------------------------
// PostInit
//-------------------------------------------------------------------------------------------------
void TheGame::PostInit()
{
	unsigned char* pixels;
	short width, height;
	int bytesPerPixel;
	char filename[ 1024 ];
	int mapLevelNumber = START_MAP_NUMBER;
	sprintf( filename, "map%02d.tga", mapLevelNumber );
	LoadTarga( filename, pixels, width, height, bytesPerPixel );
	if( bytesPerPixel != 3 )
	{
		PANIC_MSG2( "%s had unsupported bit-depth of %d bpp\n", filename, bytesPerPixel );
		return;
	}

	g_mapData = new MapData( pixels, width, height );
	g_renderer.SetBackgroundTexture( pixels, width, height );

	RENDERER->ShowFpsToggle();

	g_player = new Player;
	SpawnEnemies(); // NOTE this "new"s memory that should be deleted if new maps are loaded

	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glShadeModel( GL_SMOOTH );
}


//-------------------------------------------------------------------------------------------------
// UpdateEverything
//-------------------------------------------------------------------------------------------------
void TheGame::UpdateEverything()
{
	// update enemies
	UpdateEnemies();

	// update player
	VALIDATE( g_player );
	g_player->Move();
	g_player->UpdateBustaculation();
}


//-------------------------------------------------------------------------------------------------
// DrawEverything
//-------------------------------------------------------------------------------------------------
void TheGame::DrawEverything()
{
	/// Get the size and center of the current video mode
	Vector2 screenSize( (float) VIDEO->GetCurrentVidModeWidth(), (float) VIDEO->GetCurrentVidModeHeight() );
	Vector2 screenCenter( screenSize * 0.5f );
	Vector2 bottomLeft( 0, 960 );
	Vector2 topRight( 1280, 0 );

	/// Set up a 2d drawing coordinate system with 0.0 at the top-left
	RENDERER->SetCoordinateSystem( bottomLeft, topRight );

	/// Draw the backdrop and maze
	g_renderer.RenderBackground();

	/// Draw enemies
	g_renderer.RenderEnemies();

	/// Draw the player
	g_player->Draw();

	g_renderer.RenderBustaculationMeter();
}


//-------------------------------------------------------------------------------------------------
// DoRealTimeAlterations
//-------------------------------------------------------------------------------------------------
void TheGame::DoRealTimeAlterations()
{
	static int lastMouseX = 0;
	static int lastMouseY = 0;

	const int mouseBlockRadius = 3;

	int x,y;
	int minX = Max( MAP_MIN_X, lastMouseX - mouseBlockRadius );
	int maxX = Min( MAP_MAX_X, lastMouseX + mouseBlockRadius );
	int minY = Max( MAP_MIN_Y, lastMouseY - mouseBlockRadius );
	int maxY = Min( MAP_MAX_Y, lastMouseY + mouseBlockRadius );

	for( x = minX; x <= maxX; x ++ )
	for( y = minY; y <= maxY; y ++ )
	{
		if( g_mapData->data[ x ][ y ].type == GRIDTYPE_TEMPOCCLUDER )
		{
			g_mapData->data[ x ][ y ].type = GRIDTYPE_EMPTY;
		}
	}

	INPUT->GetMousePosClientAbsolute( lastMouseX, lastMouseY );

	minX = Max( MAP_MIN_X, lastMouseX - mouseBlockRadius );
	maxX = Min( MAP_MAX_X, lastMouseX + mouseBlockRadius );
	minY = Max( MAP_MIN_Y, lastMouseY - mouseBlockRadius );
	maxY = Min( MAP_MAX_Y, lastMouseY + mouseBlockRadius );

	for( x = minX; x <= maxX; x ++ )
	for( y = minY; y <= maxY; y ++ )
	{
		if( g_mapData->data[ x ][ y ].type == GRIDTYPE_EMPTY )
		{
			g_mapData->data[ x ][ y ].type = GRIDTYPE_TEMPOCCLUDER;
		}
	}
}


//-------------------------------------------------------------------------------------------------
// RunFrame
//-------------------------------------------------------------------------------------------------
void TheGame::RunFrame()
{
	/// Remember last mouse position
	DoRealTimeAlterations();

	/// Update stuff
	UpdateEverything();

	/// Draw stuff
	DrawEverything();
}


//-------------------------------------------------------------------------------------------------
// OnKeyDown
//-------------------------------------------------------------------------------------------------
void TheGame::OnKeyDown( const byte_t key )
{
	if( key == VK_ESCAPE )
	{
		APPLICATION->RequestShutdown();
	}
	else if( key == VK_F11 )
	{
		RENDERER->ShowFpsToggle();
	}
	else if( key == VK_TILDE )
	{
		RENDERER->ShowConsoleToggle();
	}
	else if( key == VK_RIGHT || key == VK_UP || key == VK_LEFT || key == VK_DOWN )
	{
		if( g_player )
		{
			g_player->OnKeyAction( key, true );
		}
	}
}


//-------------------------------------------------------------------------------------------------
// OnKeyUp
//-------------------------------------------------------------------------------------------------
void TheGame::OnKeyUp( const byte_t key )
{
	if( key == VK_RIGHT || key == VK_UP || key == VK_LEFT || key == VK_DOWN )
	{
		if( g_player )
		{
			g_player->OnKeyAction( key, false );
		}
	}
}


//-------------------------------------------------------------------------------------------------
// SpawnEnemies
//-------------------------------------------------------------------------------------------------
void TheGame::SpawnEnemies()
{
	/// Search for pixels of type GRIDTYPE_ENEMYSTART
	for( int y = 0; y < MAP_HEIGHT; y++ )
	for( int x = 0; x < MAP_WIDTH; x++ )
	{
		if( g_mapData->data[ x ][ y ].type == GRIDTYPE_ENEMYSTART )
		{
			/// Find the nearest enemy path point to this point
			Vector2 origin( (float)x, (float)y );
			float angle = 0;
			EnemyGuard::m_guardList.push_back( new EnemyGuard( origin, angle ) );
		}
	}
}

//-------------------------------------------------------------------------------------------------
void TheGame::UpdateEnemies()
{
	// for_each is not available for some reason
	std::list< EnemyGuard* >::iterator iter;
	std::list< EnemyGuard* >& guardList = EnemyGuard::m_guardList;
	for( iter = guardList.begin(); iter != guardList.end(); iter ++ )
	{
		(*iter)->Think();
	}
}


//-------------------------------------------------------------------------------------------------
bool TheGame::IsPointInViewCone( const Vector2& viewerOrigin, const Vector2& testOrigin, const float forwardDeg, const float fovDeg, const float viewDist )
{
	// do simple distance check first
	Vector2 delta = testOrigin - viewerOrigin;
	const double dist = delta.CalcLength();
	if( dist > (PLAYER_RADIUS + viewDist ) )
	{
		return false;
	}

	// do cone check - this is retarded code for handling cross-zero-degrees 
	const double halfFov = fovDeg / 2.0;
	double degOfDelta = delta.GetDegrees();
	double minDeg = forwardDeg - halfFov;
	double maxDeg = forwardDeg + halfFov;

	if( maxDeg > 360 )
	{
		degOfDelta += 360;
	}

	if( minDeg > 0 )
	{
		if( degOfDelta < 0 )
			degOfDelta += 360;
	}
	if( degOfDelta > minDeg && degOfDelta < maxDeg )
		return true;
	else
		return false;
}

